/* gp.c  visual guide reaching by genetic programming
 * main
 * 29 Jan 96
 * 7 Mar 96
 *
 * simplified  4 Mar 97
 * update 23 Feb 2006
 * P. Chongstitvatna
 */

#include "gp.h"

extern point shoulder, elbow, wrist, tip, target; /* absolute positions */
extern char  finishF, dieF, outF, hitF, incF, decF, seeF; /* flags */
extern int   pain;
extern int   ds,de,dw;                         /* accumulated motions */
extern int   sumdist;

int   clock, gen;
int   initialdistance, lastdistance;
int   pgm[POP], pfit[POP], ptime[POP], dist[POP];  /* current stat */
char  success[POP];
int   rank[NBEST];

#define forBest(i)       for( i=0; i<NBEST; i++ )
#define notmove(p1,p2)   ( p1.x == p2.x && p1.y == p2.y )

int   traceF = 0;

PRIVATE void runaprog(int ap)   /* simulate run a program */
{
  int d2, bore;
  point ow;

  lastdistance = initialdistance;
  sumdist = 0;
  finishF = dieF = outF = 0;
  clock = 0; bore = pain = 0;
  while( clock++ < TIMELIMIT && ! finishF && ! dieF ) {
    ow = wrist;
    ds = de = dw = 0;
    evalprog(ap);
    gmove(ds,de,dw);

    if( traceF ){
      armpos();
	  printf("%d %d %d %d %d %d\n",
	    elbow.x, elbow.y, wrist.x, wrist.y, tip.x, tip.y);
	}

    d2 = distance(tip,target);
    incF = (d2 > lastdistance);
    decF = (d2 < lastdistance);
    lastdistance = d2;
    hitF = (pain > 0);
    seeF = ! fastscan(tip,target);
    if( notmove(ow,wrist) ) bore++; else { bore = 0; ow = wrist; }
    if( bore > BORELIMIT ) dieF = 1;
    if( pain > PAINLIMIT ) dieF = 2;
  }
}

/* return the fitness of current prog at the end of sim */
PRIVATE int evalfitness(void )
{
  double h;

  h = -2000*(double)lastdistance/initialdistance;
  h -= 100*(double)sumdist/initialdistance;
  if ( ! seeF ) h -= 1000;
  if ( dieF ) h -= 4000;
  return (int) h;
}

/*  Took algorithm, pick min of the first NBEST and swap with the rest.
 *  This is 2 times faster than straight forward pick max NBEST times.
 *  When finish, pfit[] is substitute in place,
 *  pfit[] will be the NBEST and index is stored
 *  in rank[].  Use this index to get other values (pgm[],ptime[]...)
 */
PRIVATE void Tbest(void)
{
  int i,p,min,mi;

  forBest(i) rank[i] = i;
  p = NBEST;
  while( p < POP ) {
    for( min = pfit[0], mi = 0, i=1; i<NBEST; i++ )
      if( pfit[i] < min ) {
		min = pfit[i];
		mi = i;
      }
    while( p < POP ) {
      if( pfit[p] > min ) {
		pfit[mi] = pfit[p];
		rank[mi] = p;
		p++;
		break;
      }
      p++;
    }
  }
}

/* move the stat to the correct places and copy prog to new space
 * reclaim the old area, (garbage collect)
 */
PRIVATE void keepbest(void)
{
  int i, t[NBEST]; /* t[] is temp area */

  Tbest();
  forBest(i) t[i] = pgm[rank[i]];
  forBest(i) pgm[i] = t[i];
  forBest(i) t[i] = ptime[rank[i]];
  forBest(i) ptime[i] = t[i];          /* pfit[] is already in place */
  forBest(i) t[i] = success[rank[i]];
  forBest(i) success[i] = t[i];
  forBest(i) t[i] = dist[rank[i]];
  forBest(i) dist[i] = t[i];

  markcell();
  forBest(i) pgm[i] = copyprog(pgm[i]);
  reclaim();
}

/* pick parents with replacement but parents must be diff */
PRIVATE void docrossover(void)
{
  int k,ap1,ap2,r,s;

  for( k=40; k<200; k+=2 ) {
    r = myrnd(NBEST);
    do { s = myrnd(NBEST); } while( s == r );
    ap1 = copyprog(pgm[r]);
    ap2 = copyprog(pgm[s]);
    crossover(ap1,ap2);
    pgm[k] = ap1;
    pgm[k+1] = ap2;
  }
}

PRIVATE void initprog(void)
{
  int i,k;
  for( i=0; i<POP; i++) pgm[i] = genprogn(PLEN,&k);
}

PRIVATE void stat2(int i) /* for one pgm[i] */
{
   printf("pgm %d len %d fit %d time %d success %d dist %d\n",
	     i,proglen(pgm[i]),pfit[i],ptime[i],success[i],dist[i]);
   printf(": "); tofile(pgm[i],stdout); nl();
}

PRIVATE void stat1(void)  /* report only the NBEST */
{
  int i,sl,sf,max,mi;

  sl = sf = 0; max = pfit[0]; mi = 0;
  forBest(i) {
    sl += proglen(pgm[i]);
    sf += pfit[i];
    if( max < pfit[i] ) { max = pfit[i]; mi = i; }
	if ( gen == GEN ) stat2(i);		/* report individual of the last gen */
  }
  printf("gen %d best fit %d av fit %d av len %d\n",
	 gen,max,sf/NBEST,sl/NBEST);
}

/* run experiment for gmax generations */
PRIVATE void runexperiment(int gmax)
{
  int i,k;

  for(gen=1; gen<=gmax; gen++){
    k = ( gen == 1 ) ? 0 : NBEST;
    for( i=k; i<POP; i++ ) {
      sethome();
      runaprog( pgm[i] );
      pfit[i] = evalfitness();
      ptime[i] = clock;
      success[i] = finishF;
      dist[i] = sumdist;
    }                         /* end one generation */
    keepbest();
    stat1();
    docrossover();
    for(i=0; i<100; i++) pgm[i+200] = addprog( copyprog(pgm[i]));
    for(i=0; i<200; i++) pgm[i+300] = extendprog( copyprog(pgm[i]));
  }
}
/* trace joint angle, input prog from a file */
/* each prog ends with exactly one newline */
PRIVATE void traceprog(char *fname)
{
  FILE *fp;
  int i,k,c;

  fp = fopen(fname,"r");
  initcell();
  i = 0;
  c = fromfile(fp);
  while(c > 0){
    pgm[i] = c;
	fgetc(fp);					// read nl
//	tofile(pgm[i],stdout); nl();
    c = fromfile(fp);
    i++;
  }
  fclose(fp);
  printf("no of prog %d\n",i);
  env0();
  sethome();
  initialdistance = distance(tip,target);
  traceF = 1;
  // run only the first individual
  k = 0;
  sethome();
  runaprog(pgm[k]);
  printf("pgm %d finishF %d dieF %d seeF %d pain %d ",
	 k,finishF,dieF,seeF,pain);
  printf("fit %d clock %d len %d dist %d\n",
     evalfitness(),clock,proglen(pgm[k]),sumdist);
}

int main(void)
{
//  traceprog("prog1.txt");
//  exit(0);
  mysrand(1);
  initcell();
  initprog();
  env0();
  sethome();
  initialdistance = distance(tip,target);
  runexperiment(GEN);
}

/* *****      Utilities *****

dumptofile()
{
  int i;
  for( i=0; i<100; i+=10 ) { tofile(pgm[i],stdout); nl(); }
}

dumpenv()
{
  int x,y;
  for( y=0; y<IMH; y++ )
    for( x=0; x<IMW; x++ )
      putchar(im[x][y]);
}

testeqprog(a,b)
int a,b;
{
  int i,j;
  for( i=a; i<=b; i++ )
    for( j=i+1; j<=b; j++ )
      if ( eqprog(pgm[i],pgm[j]) ) printf("eq pgm %d %d\n",i,j);
}

--* stat of the whole population
 * dump individual stat and program of NBEST, best, worst that success *--
PRIVATE stat()
{
  int i,k,sum,suc,sumf, max,min,mx,mn;
  int s[NBEST], scnt;

  for( sum=0, suc=0, sumf = 0, i=0; i<POP; i++ ) {
    sum += proglen(pgm[i]);
    suc += success[i];
    sumf += pfit[i];
  }
  printf("Last Gen : total success %d av len %d av fit %d\n",
	 suc,sum/POP,sumf/POP);
  if( suc == 0 ) {
    printf("the first ten programs\n");
    for( i=0; i<10; i++ ) { printf("x "); tofile(pgm[i],stdout); nl(); }
    return;
  }
  scnt = 0;
  forBest(i) if( success[i] ) { s[scnt++] = i; stat2(i); }
  mx = mn = s[0];
  max = min = pfit[mx];
  for(  i=1; i<scnt; i++ ) {
    k = s[i];
    if( max < pfit[k] ) { max = pfit[k]; mx = k; }
    if( min > pfit[k] ) { min = pfit[k]; mn = k; }
  }
  for( i=0; i<scnt; i++ ) { printf("x "); tofile(pgm[s[i]],stdout); nl(); }
  printf("best "); stat2(mx);
  printf("worst "); stat2(mn);
  printf("b "); tofile(pgm[mx],stdout); nl();
  printf("w "); tofile(pgm[mn],stdout); nl();
}

--*  trace all programs from the file fname, last program must end with '.' *--
PRIVATE traceprog(fname)
char *fname;
{
  FILE *fp;
  int i,k,c;

  fp = fopen(fname,"r");
  i = 0;
  do { pgm[i++] = fromfile(fp); } while ( (c = fgetc(fp)) != '.' );
  for( k=0; k<i; k++ ) {
    sethome();
    if( logF ) where2();
    runaprog(pgm[k]);
    if( ! logF ) {
      printf("pgm %d finishF %d dieF %d seeF %d pain %d ",
	 k,finishF,dieF,seeF,pain);
      printf("fit %d clock %d len %d dist %d\n",
	 evalfitness(),clock,proglen(pgm[k]),sumdist);
    }
  }
  close(fp);
}

***** */
